@@ -5,13 +5,21 @@
  * file, You can obtain one at http://mozilla.org/MPL/2.0/. */
 
 #include "ImageContainer.h"
-#include <string.h>    // for memcpy, memset
-#include "GLImages.h"  // for SurfaceTextureImage
+
+#include <string.h>  // for memcpy, memset
+
+#include "GLImages.h"    // for SurfaceTextureImage
+#include "YCbCrUtils.h"  // for YCbCr conversions
 #include "gfx2DGlue.h"
 #include "gfxPlatform.h"  // for gfxPlatform
 #include "gfxUtils.h"     // for gfxUtils
 #include "libyuv.h"
-#include "mozilla/RefPtr.h"                 // for already_AddRefed
+#include "mozilla/CheckedInt.h"
+#include "mozilla/ProfilerLabels.h"
+#include "mozilla/RefPtr.h"  // for already_AddRefed
+#include "mozilla/StaticPrefs_layers.h"
+#include "mozilla/gfx/2D.h"
+#include "mozilla/gfx/gfxVars.h"
 #include "mozilla/ipc/CrossProcessMutex.h"  // for CrossProcessMutex, etc
 #include "mozilla/layers/CompositorTypes.h"
 #include "mozilla/layers/ImageBridgeChild.h"     // for ImageBridgeChild
@@ -19,29 +27,26 @@
 #include "mozilla/layers/ImageDataSerializer.h"  // for SurfaceDescriptorBuffer
 #include "mozilla/layers/LayersMessages.h"
 #include "mozilla/layers/SharedPlanarYCbCrImage.h"
-#include "mozilla/layers/SharedSurfacesChild.h"  // for SharedSurfacesAnimation
 #include "mozilla/layers/SharedRGBImage.h"
 #include "mozilla/layers/TextureClientRecycleAllocator.h"
-#include "mozilla/gfx/gfxVars.h"
+#include "nsProxyRelease.h"
 #include "nsISupportsUtils.h"  // for NS_IF_ADDREF
-#include "YCbCrUtils.h"        // for YCbCr conversions
-#include "gfx2DGlue.h"
-#include "mozilla/gfx/2D.h"
-#include "mozilla/CheckedInt.h"
 
 #ifdef XP_MACOSX
+#  include "MacIOSurfaceImage.h"
 #  include "mozilla/gfx/QuartzSupport.h"
 #endif
 
 #ifdef XP_WIN
-#  include "gfxWindowsPlatform.h"
 #  include <d3d10_1.h>
+
+#  include "gfxWindowsPlatform.h"
 #  include "mozilla/gfx/DeviceManagerDx.h"
+#  include "mozilla/layers/D3D11ShareHandleImage.h"
 #  include "mozilla/layers/D3D11YCbCrImage.h"
 #endif
 
-namespace mozilla {
-namespace layers {
+namespace mozilla::layers {
 
 using namespace mozilla::gfx;
 using namespace mozilla::ipc;
@@ -85,10 +90,7 @@ UniquePtr<uint8_t[]> BufferRecycleBin::GetBuffer(uint32_t aSize) {
     return UniquePtr<uint8_t[]>(new (fallible) uint8_t[aSize]);
   }
 
-  uint32_t last = mRecycledBuffers.Length() - 1;
-  UniquePtr<uint8_t[]> result = std::move(mRecycledBuffers[last]);
-  mRecycledBuffers.RemoveElementAt(last);
-  return result;
+  return mRecycledBuffers.PopLastElement();
 }
 
 void BufferRecycleBin::ClearRecycledBuffers() {
@@ -103,7 +105,7 @@ ImageContainerListener::ImageContainerListener(ImageContainer* aImageContainer)
     : mLock("mozilla.layers.ImageContainerListener.mLock"),
       mImageContainer(aImageContainer) {}
 
-ImageContainerListener::~ImageContainerListener() {}
+ImageContainerListener::~ImageContainerListener() = default;
 
 void ImageContainerListener::NotifyComposite(
     const ImageCompositeNotification& aNotification) {
@@ -173,13 +175,6 @@ void ImageContainer::EnsureImageClient() {
   }
 }
 
-SharedSurfacesAnimation* ImageContainer::EnsureSharedSurfacesAnimation() {
-  if (!mSharedAnimation) {
-    mSharedAnimation = new SharedSurfacesAnimation();
-  }
-  return mSharedAnimation;
-}
-
 ImageContainer::ImageContainer(Mode flag)
     : mRecursiveMutex("ImageContainer.mRecursiveMutex"),
       mGenerationCounter(++sGenerationCounter),
@@ -218,9 +213,33 @@ ImageContainer::~ImageContainer() {
       imageBridge->ForgetImageContainer(mAsyncContainerHandle);
     }
   }
-  if (mSharedAnimation) {
-    mSharedAnimation->Destroy();
+}
+
+Maybe<SurfaceDescriptor> Image::GetDesc() { return GetDescFromTexClient(); }
+
+Maybe<SurfaceDescriptor> Image::GetDescFromTexClient(
+    TextureClient* const tcOverride) {
+  RefPtr<TextureClient> tc = tcOverride;
+  if (!tcOverride) {
+    tc = GetTextureClient(nullptr);
   }
+  if (!tc) {
+    return {};
+  }
+
+  const auto& tcd = tc->GetInternalData();
+
+  SurfaceDescriptor ret;
+  if (!tcd->Serialize(ret)) {
+    return {};
+  }
+  return Some(ret);
+}
+
+already_AddRefed<ImageContainerListener>
+ImageContainer::GetImageContainerListener() const {
+  MOZ_ASSERT(InImageBridgeChildThread());
+  return do_AddRef(mNotifyCompositeListener);
 }
 
 RefPtr<PlanarYCbCrImage> ImageContainer::CreatePlanarYCbCrImage() {
@@ -229,16 +248,22 @@ RefPtr<PlanarYCbCrImage> ImageContainer::CreatePlanarYCbCrImage() {
   if (mImageClient && mImageClient->AsImageClientSingle()) {
     return new SharedPlanarYCbCrImage(mImageClient);
   }
+  if (mRecycleAllocator) {
+    return new SharedPlanarYCbCrImage(mRecycleAllocator);
+  }
   return mImageFactory->CreatePlanarYCbCrImage(mScaleHint, mRecycleBin);
 }
 
 RefPtr<SharedRGBImage> ImageContainer::CreateSharedRGBImage() {
   RecursiveMutexAutoLock lock(mRecursiveMutex);
   EnsureImageClient();
-  if (!mImageClient || !mImageClient->AsImageClientSingle()) {
-    return nullptr;
+  if (mImageClient && mImageClient->AsImageClientSingle()) {
+    return new SharedRGBImage(mImageClient);
   }
-  return new SharedRGBImage(mImageClient);
+  if (mRecycleAllocator) {
+    return new SharedRGBImage(mRecycleAllocator);
+  }
+  return nullptr;
 }
 
 void ImageContainer::SetCurrentImageInternal(
@@ -285,7 +310,7 @@ void ImageContainer::SetCurrentImageInternal(
     }
   }
 
-  mCurrentImages.SwapElements(newImages);
+  mCurrentImages = std::move(newImages);
 }
 
 void ImageContainer::ClearImagesFromImageBridge() {
@@ -294,6 +319,7 @@ void ImageContainer::ClearImagesFromImageBridge() {
 }
 
 void ImageContainer::SetCurrentImages(const nsTArray<NonOwningImage>& aImages) {
+  AUTO_PROFILER_LABEL("ImageContainer::SetCurrentImages", GRAPHICS);
   MOZ_ASSERT(!aImages.IsEmpty());
   RecursiveMutexAutoLock lock(mRecursiveMutex);
   if (mIsAsync) {
@@ -306,18 +332,22 @@ void ImageContainer::SetCurrentImages(const nsTArray<NonOwningImage>& aImages) {
 }
 
 void ImageContainer::ClearAllImages() {
+  mRecursiveMutex.Lock();
   if (mImageClient) {
+    RefPtr<ImageClient> imageClient = mImageClient;
+    mRecursiveMutex.Unlock();
+
     // Let ImageClient release all TextureClients. This doesn't return
     // until ImageBridge has called ClearCurrentImageFromImageBridge.
     if (RefPtr<ImageBridgeChild> imageBridge =
             ImageBridgeChild::GetSingleton()) {
-      imageBridge->FlushAllImages(mImageClient, this);
+      imageBridge->FlushAllImages(imageClient, this);
     }
     return;
   }
 
-  RecursiveMutexAutoLock lock(mRecursiveMutex);
   SetCurrentImageInternal(nsTArray<NonOwningImage>());
+  mRecursiveMutex.Unlock();
 }
 
 void ImageContainer::ClearCachedResources() {
@@ -341,7 +371,7 @@ void ImageContainer::SetCurrentImageInTransaction(Image* aImage) {
 void ImageContainer::SetCurrentImagesInTransaction(
     const nsTArray<NonOwningImage>& aImages) {
   NS_ASSERTION(NS_IsMainThread(), "Should be on main thread.");
-  NS_ASSERTION(!mImageClient,
+  NS_ASSERTION(!HasImageClient(),
                "Should use async image transfer with ImageBridge.");
 
   SetCurrentImageInternal(aImages);
@@ -352,8 +382,8 @@ bool ImageContainer::IsAsync() const { return mIsAsync; }
 CompositableHandle ImageContainer::GetAsyncContainerHandle() {
   NS_ASSERTION(IsAsync(),
                "Shared image ID is only relevant to async ImageContainers");
-  NS_ASSERTION(mAsyncContainerHandle, "Should have a shared image ID");
   RecursiveMutexAutoLock mon(mRecursiveMutex);
+  NS_ASSERTION(mAsyncContainerHandle, "Should have a shared image ID");
   EnsureImageClient();
   return mAsyncContainerHandle;
 }
@@ -368,7 +398,7 @@ void ImageContainer::GetCurrentImages(nsTArray<OwningImage>* aImages,
                                       uint32_t* aGenerationCounter) {
   RecursiveMutexAutoLock lock(mRecursiveMutex);
 
-  *aImages = mCurrentImages;
+  *aImages = mCurrentImages.Clone();
   if (aGenerationCounter) {
     *aGenerationCounter = mGenerationCounter;
   }
@@ -411,50 +441,206 @@ void ImageContainer::NotifyDropped(uint32_t aDropped) {
   mDroppedImageCount += aDropped;
 }
 
-#ifdef XP_WIN
-D3D11YCbCrRecycleAllocator* ImageContainer::GetD3D11YCbCrRecycleAllocator(
-    KnowsCompositor* aAllocator) {
-  if (mD3D11YCbCrRecycleAllocator &&
-      aAllocator == mD3D11YCbCrRecycleAllocator->GetAllocator()) {
-    return mD3D11YCbCrRecycleAllocator;
+void ImageContainer::EnsureRecycleAllocatorForRDD(
+    KnowsCompositor* aKnowsCompositor) {
+  MOZ_ASSERT(!mIsAsync);
+  MOZ_ASSERT(XRE_IsRDDProcess());
+
+  RecursiveMutexAutoLock lock(mRecursiveMutex);
+  MOZ_ASSERT(!mImageClient);
+
+  if (mRecycleAllocator &&
+      aKnowsCompositor == mRecycleAllocator->GetKnowsCompositor()) {
+    return;
   }
 
-  if (!aAllocator->SupportsD3D11() ||
+  if (!StaticPrefs::layers_recycle_allocator_rdd_AtStartup()) {
+    return;
+  }
+
+  static const uint32_t MAX_POOLED_VIDEO_COUNT = 5;
+
+  mRecycleAllocator =
+      new layers::TextureClientRecycleAllocator(aKnowsCompositor);
+  mRecycleAllocator->SetMaxPoolSize(MAX_POOLED_VIDEO_COUNT);
+}
+
+#ifdef XP_WIN
+already_AddRefed<D3D11RecycleAllocator>
+ImageContainer::GetD3D11RecycleAllocator(KnowsCompositor* aKnowsCompositor,
+                                         gfx::SurfaceFormat aPreferredFormat) {
+  MOZ_ASSERT(aKnowsCompositor);
+
+  if (!aKnowsCompositor->SupportsD3D11()) {
+    return nullptr;
+  }
+
+  RefPtr<ID3D11Device> device = gfx::DeviceManagerDx::Get()->GetImageDevice();
+  if (!device) {
+    return nullptr;
+  }
+
+  RecursiveMutexAutoLock lock(mRecursiveMutex);
+  if (mD3D11RecycleAllocator && mD3D11RecycleAllocator->mDevice == device &&
+      mD3D11RecycleAllocator->GetKnowsCompositor() == aKnowsCompositor) {
+    return do_AddRef(mD3D11RecycleAllocator);
+  }
+
+  mD3D11RecycleAllocator =
+      new D3D11RecycleAllocator(aKnowsCompositor, device, aPreferredFormat);
+
+  if (device != DeviceManagerDx::Get()->GetCompositorDevice()) {
+    RefPtr<SyncObjectClient> syncObject =
+        SyncObjectClient::CreateSyncObjectClient(
+            aKnowsCompositor->GetTextureFactoryIdentifier().mSyncHandle,
+            device);
+    mD3D11RecycleAllocator->SetSyncObject(syncObject);
+  }
+
+  return do_AddRef(mD3D11RecycleAllocator);
+}
+
+already_AddRefed<D3D11YCbCrRecycleAllocator>
+ImageContainer::GetD3D11YCbCrRecycleAllocator(
+    KnowsCompositor* aKnowsCompositor) {
+  if (!aKnowsCompositor->SupportsD3D11() ||
       !gfx::DeviceManagerDx::Get()->GetImageDevice()) {
     return nullptr;
   }
 
-  mD3D11YCbCrRecycleAllocator = new D3D11YCbCrRecycleAllocator(aAllocator);
-  return mD3D11YCbCrRecycleAllocator;
+  RecursiveMutexAutoLock lock(mRecursiveMutex);
+  if (mD3D11YCbCrRecycleAllocator &&
+      aKnowsCompositor == mD3D11YCbCrRecycleAllocator->GetKnowsCompositor()) {
+    return do_AddRef(mD3D11YCbCrRecycleAllocator);
+  }
+
+  mD3D11YCbCrRecycleAllocator =
+      new D3D11YCbCrRecycleAllocator(aKnowsCompositor);
+  return do_AddRef(mD3D11YCbCrRecycleAllocator);
 }
 #endif
 
+#ifdef XP_MACOSX
+already_AddRefed<MacIOSurfaceRecycleAllocator>
+ImageContainer::GetMacIOSurfaceRecycleAllocator() {
+  RecursiveMutexAutoLock lock(mRecursiveMutex);
+  if (!mMacIOSurfaceRecycleAllocator) {
+    mMacIOSurfaceRecycleAllocator = new MacIOSurfaceRecycleAllocator();
+  }
+
+  return do_AddRef(mMacIOSurfaceRecycleAllocator);
+}
+#endif
+
+// -
+// https://searchfox.org/mozilla-central/source/dom/media/ipc/RemoteImageHolder.cpp#46
+
+Maybe<PlanarYCbCrData> PlanarYCbCrData::From(
+    const SurfaceDescriptorBuffer& sdb) {
+  if (sdb.desc().type() != BufferDescriptor::TYCbCrDescriptor) {
+    return {};
+  }
+  const YCbCrDescriptor& yuvDesc = sdb.desc().get_YCbCrDescriptor();
+
+  Maybe<Range<uint8_t>> buffer;
+  const MemoryOrShmem& memOrShmem = sdb.data();
+  switch (memOrShmem.type()) {
+    case MemoryOrShmem::Tuintptr_t:
+      gfxCriticalError() << "PlanarYCbCrData::From SurfaceDescriptorBuffer "
+                            "w/uintptr_t unsupported.";
+      break;
+    case MemoryOrShmem::TShmem:
+      buffer.emplace(memOrShmem.get_Shmem().Range<uint8_t>());
+      break;
+    default:
+      MOZ_ASSERT(false, "Unknown MemoryOrShmem type");
+      break;
+  }
+  if (!buffer) {
+    return {};
+  }
+
+  PlanarYCbCrData yuvData;
+  yuvData.mYStride = AssertedCast<int32_t>(yuvDesc.yStride());
+  yuvData.mCbCrStride = AssertedCast<int32_t>(yuvDesc.cbCrStride());
+  // default mYSkip, mCbSkip, mCrSkip because not held in YCbCrDescriptor
+  yuvData.mYSkip = yuvData.mCbSkip = yuvData.mCrSkip = 0;
+  yuvData.mPictureRect = yuvDesc.display();
+  yuvData.mStereoMode = yuvDesc.stereoMode();
+  yuvData.mColorDepth = yuvDesc.colorDepth();
+  yuvData.mYUVColorSpace = yuvDesc.yUVColorSpace();
+  yuvData.mColorRange = yuvDesc.colorRange();
+  yuvData.mChromaSubsampling = yuvDesc.chromaSubsampling();
+
+  const auto GetPlanePtr = [&](const uint32_t beginOffset,
+                               const gfx::IntSize size,
+                               const int32_t stride) -> uint8_t* {
+    if (size.width > stride) return nullptr;
+    auto bytesNeeded = CheckedInt<uintptr_t>(stride) *
+                       size.height;  // Don't accept `stride*(h-1)+w`.
+    bytesNeeded += beginOffset;
+    if (!bytesNeeded.isValid() || bytesNeeded.value() > buffer->length()) {
+      gfxCriticalError()
+          << "PlanarYCbCrData::From asked for out-of-bounds plane data.";
+      return nullptr;
+    }
+    return (buffer->begin() + beginOffset).get();
+  };
+  yuvData.mYChannel =
+      GetPlanePtr(yuvDesc.yOffset(), yuvDesc.ySize(), yuvData.mYStride);
+  yuvData.mCbChannel =
+      GetPlanePtr(yuvDesc.cbOffset(), yuvDesc.cbCrSize(), yuvData.mCbCrStride);
+  yuvData.mCrChannel =
+      GetPlanePtr(yuvDesc.crOffset(), yuvDesc.cbCrSize(), yuvData.mCbCrStride);
+
+  if (yuvData.mYSkip || yuvData.mCbSkip || yuvData.mCrSkip ||
+      yuvDesc.ySize().width < 0 || yuvDesc.ySize().height < 0 ||
+      yuvDesc.cbCrSize().width < 0 || yuvDesc.cbCrSize().height < 0 ||
+      yuvData.mYStride < 0 || yuvData.mCbCrStride < 0 || !yuvData.mYChannel ||
+      !yuvData.mCbChannel || !yuvData.mCrChannel) {
+    gfxCriticalError() << "Unusual PlanarYCbCrData: " << yuvData.mYSkip << ","
+                       << yuvData.mCbSkip << "," << yuvData.mCrSkip << ", "
+                       << yuvDesc.ySize().width << "," << yuvDesc.ySize().height
+                       << ", " << yuvDesc.cbCrSize().width << ","
+                       << yuvDesc.cbCrSize().height << ", " << yuvData.mYStride
+                       << "," << yuvData.mCbCrStride << ", "
+                       << yuvData.mYChannel << "," << yuvData.mCbChannel << ","
+                       << yuvData.mCrChannel;
+    return {};
+  }
+
+  return Some(yuvData);
+}
+
+// -
+
 PlanarYCbCrImage::PlanarYCbCrImage()
     : Image(nullptr, ImageFormat::PLANAR_YCBCR),
       mOffscreenFormat(SurfaceFormat::UNKNOWN),
       mBufferSize(0) {}
 
 nsresult PlanarYCbCrImage::BuildSurfaceDescriptorBuffer(
-    SurfaceDescriptorBuffer& aSdBuffer) {
+    SurfaceDescriptorBuffer& aSdBuffer,
+    const std::function<MemoryOrShmem(uint32_t)>& aAllocate) {
   const PlanarYCbCrData* pdata = GetData();
   MOZ_ASSERT(pdata, "must have PlanarYCbCrData");
   MOZ_ASSERT(pdata->mYSkip == 0 && pdata->mCbSkip == 0 && pdata->mCrSkip == 0,
              "YCbCrDescriptor doesn't hold skip values");
-  MOZ_ASSERT(pdata->mPicX == 0 && pdata->mPicY == 0,
-             "YCbCrDescriptor doesn't hold picx or picy");
 
+  auto ySize = pdata->YDataSize();
+  auto cbcrSize = pdata->CbCrDataSize();
   uint32_t yOffset;
   uint32_t cbOffset;
   uint32_t crOffset;
-  ImageDataSerializer::ComputeYCbCrOffsets(
-      pdata->mYStride, pdata->mYSize.height, pdata->mCbCrStride,
-      pdata->mCbCrSize.height, yOffset, cbOffset, crOffset);
+  ImageDataSerializer::ComputeYCbCrOffsets(pdata->mYStride, ySize.height,
+                                           pdata->mCbCrStride, cbcrSize.height,
+                                           yOffset, cbOffset, crOffset);
 
-  aSdBuffer.desc() = YCbCrDescriptor(
-      pdata->mYSize, pdata->mYStride, pdata->mCbCrSize, pdata->mCbCrStride,
-      yOffset, cbOffset, crOffset, pdata->mStereoMode, pdata->mColorDepth,
-      pdata->mYUVColorSpace,
-      /*hasIntermediateBuffer*/ false);
+  uint32_t bufferSize = ImageDataSerializer::ComputeYCbCrBufferSize(
+      ySize, pdata->mYStride, cbcrSize, pdata->mCbCrStride, yOffset, cbOffset,
+      crOffset);
+
+  aSdBuffer.data() = aAllocate(bufferSize);
 
   uint8_t* buffer = nullptr;
   const MemoryOrShmem& memOrShmem = aSdBuffer.data();
@@ -466,19 +652,24 @@ nsresult PlanarYCbCrImage::BuildSurfaceDescriptorBuffer(
       buffer = memOrShmem.get_Shmem().get<uint8_t>();
       break;
     default:
-      MOZ_ASSERT(false, "Unknown MemoryOrShmem type");
+      buffer = nullptr;
+      break;
   }
-  MOZ_ASSERT(buffer, "no valid buffer available to copy image data");
   if (!buffer) {
-    return NS_ERROR_INVALID_ARG;
+    return NS_ERROR_OUT_OF_MEMORY;
   }
 
-  CopyPlane(buffer + yOffset, pdata->mYChannel, pdata->mYSize, pdata->mYStride,
+  aSdBuffer.desc() = YCbCrDescriptor(
+      pdata->mPictureRect, ySize, pdata->mYStride, cbcrSize, pdata->mCbCrStride,
+      yOffset, cbOffset, crOffset, pdata->mStereoMode, pdata->mColorDepth,
+      pdata->mYUVColorSpace, pdata->mColorRange, pdata->mChromaSubsampling);
+
+  CopyPlane(buffer + yOffset, pdata->mYChannel, ySize, pdata->mYStride,
             pdata->mYSkip);
-  CopyPlane(buffer + cbOffset, pdata->mCbChannel, pdata->mCbCrSize,
-            pdata->mCbCrStride, pdata->mCbSkip);
-  CopyPlane(buffer + crOffset, pdata->mCrChannel, pdata->mCbCrSize,
-            pdata->mCbCrStride, pdata->mCrSkip);
+  CopyPlane(buffer + cbOffset, pdata->mCbChannel, cbcrSize, pdata->mCbCrStride,
+            pdata->mCbSkip);
+  CopyPlane(buffer + crOffset, pdata->mCrChannel, cbcrSize, pdata->mCbCrStride,
+            pdata->mCrSkip);
   return NS_OK;
 }
 
@@ -539,9 +730,12 @@ static void CopyPlane(uint8_t* aDst, const uint8_t* aSrc,
 bool RecyclingPlanarYCbCrImage::CopyData(const Data& aData) {
   // update buffer size
   // Use uint32_t throughout to match AllocateBuffer's param and mBufferSize
+  auto ySize = aData.YDataSize();
+  auto cbcrSize = aData.CbCrDataSize();
   const auto checkedSize =
-      CheckedInt<uint32_t>(aData.mCbCrStride) * aData.mCbCrSize.height * 2 +
-      CheckedInt<uint32_t>(aData.mYStride) * aData.mYSize.height;
+      CheckedInt<uint32_t>(aData.mCbCrStride) * cbcrSize.height * 2 +
+      CheckedInt<uint32_t>(aData.mYStride) * ySize.height *
+          (aData.mAlpha ? 2 : 1);
 
   if (!checkedSize.isValid()) return false;
 
@@ -554,22 +748,25 @@ bool RecyclingPlanarYCbCrImage::CopyData(const Data& aData) {
   // update buffer size
   mBufferSize = size;
 
-  mData = aData;
+  mData = aData;  // mAlpha will be set if aData has it
   mData.mYChannel = mBuffer.get();
-  mData.mCbChannel = mData.mYChannel + mData.mYStride * mData.mYSize.height;
-  mData.mCrChannel =
-      mData.mCbChannel + mData.mCbCrStride * mData.mCbCrSize.height;
+  mData.mCbChannel = mData.mYChannel + mData.mYStride * ySize.height;
+  mData.mCrChannel = mData.mCbChannel + mData.mCbCrStride * cbcrSize.height;
   mData.mYSkip = mData.mCbSkip = mData.mCrSkip = 0;
 
-  CopyPlane(mData.mYChannel, aData.mYChannel, aData.mYSize, aData.mYStride,
+  CopyPlane(mData.mYChannel, aData.mYChannel, ySize, aData.mYStride,
             aData.mYSkip);
-  CopyPlane(mData.mCbChannel, aData.mCbChannel, aData.mCbCrSize,
-            aData.mCbCrStride, aData.mCbSkip);
-  CopyPlane(mData.mCrChannel, aData.mCrChannel, aData.mCbCrSize,
-            aData.mCbCrStride, aData.mCrSkip);
+  CopyPlane(mData.mCbChannel, aData.mCbChannel, cbcrSize, aData.mCbCrStride,
+            aData.mCbSkip);
+  CopyPlane(mData.mCrChannel, aData.mCrChannel, cbcrSize, aData.mCbCrStride,
+            aData.mCrSkip);
+  if (aData.mAlpha) {
+    CopyPlane(mData.mAlpha->mChannel, aData.mAlpha->mChannel, ySize,
+              aData.mYStride, aData.mYSkip);
+  }
 
-  mSize = aData.mPicSize;
-  mOrigin = gfx::IntPoint(aData.mPicX, aData.mPicY);
+  mSize = aData.mPictureRect.Size();
+  mOrigin = aData.mPictureRect.TopLeft();
   return true;
 }
 
@@ -580,8 +777,8 @@ gfxImageFormat PlanarYCbCrImage::GetOffscreenFormat() const {
 
 bool PlanarYCbCrImage::AdoptData(const Data& aData) {
   mData = aData;
-  mSize = aData.mPicSize;
-  mOrigin = gfx::IntPoint(aData.mPicX, aData.mPicY);
+  mSize = aData.mPictureRect.Size();
+  mOrigin = aData.mPictureRect.TopLeft();
   return true;
 }
 
@@ -620,13 +817,20 @@ already_AddRefed<gfx::SourceSurface> PlanarYCbCrImage::GetAsSourceSurface() {
   return surface.forget();
 }
 
+PlanarYCbCrImage::~PlanarYCbCrImage() {
+  NS_ReleaseOnMainThread("PlanarYCbCrImage::mSourceSurface",
+                         mSourceSurface.forget());
+}
+
 NVImage::NVImage() : Image(nullptr, ImageFormat::NV_IMAGE), mBufferSize(0) {}
 
-NVImage::~NVImage() = default;
+NVImage::~NVImage() {
+  NS_ReleaseOnMainThread("NVImage::mSourceSurface", mSourceSurface.forget());
+}
 
 IntSize NVImage::GetSize() const { return mSize; }
 
-IntRect NVImage::GetPictureRect() const { return mData.GetPictureRect(); }
+IntRect NVImage::GetPictureRect() const { return mData.mPictureRect; }
 
 already_AddRefed<SourceSurface> NVImage::GetAsSourceSurface() {
   if (mSourceSurface) {
@@ -636,31 +840,30 @@ already_AddRefed<SourceSurface> NVImage::GetAsSourceSurface() {
 
   // Convert the current NV12 or NV21 data to YUV420P so that we can follow the
   // logics in PlanarYCbCrImage::GetAsSourceSurface().
-  const int bufferLength = mData.mYSize.height * mData.mYStride +
-                           mData.mCbCrSize.height * mData.mCbCrSize.width * 2;
-  auto* buffer = new uint8_t[bufferLength];
+  auto ySize = mData.YDataSize();
+  auto cbcrSize = mData.CbCrDataSize();
+  const int bufferLength =
+      ySize.height * mData.mYStride + cbcrSize.height * cbcrSize.width * 2;
+  UniquePtr<uint8_t[]> buffer(new uint8_t[bufferLength]);
 
   Data aData = mData;
-  aData.mCbCrStride = aData.mCbCrSize.width;
+  aData.mCbCrStride = cbcrSize.width;
   aData.mCbSkip = 0;
   aData.mCrSkip = 0;
-  aData.mYChannel = buffer;
-  aData.mCbChannel = aData.mYChannel + aData.mYSize.height * aData.mYStride;
-  aData.mCrChannel =
-      aData.mCbChannel + aData.mCbCrSize.height * aData.mCbCrStride;
+  aData.mYChannel = buffer.get();
+  aData.mCbChannel = aData.mYChannel + ySize.height * aData.mYStride;
+  aData.mCrChannel = aData.mCbChannel + cbcrSize.height * aData.mCbCrStride;
 
   if (mData.mCbChannel < mData.mCrChannel) {  // NV12
     libyuv::NV12ToI420(mData.mYChannel, mData.mYStride, mData.mCbChannel,
                        mData.mCbCrStride, aData.mYChannel, aData.mYStride,
                        aData.mCbChannel, aData.mCbCrStride, aData.mCrChannel,
-                       aData.mCbCrStride, aData.mYSize.width,
-                       aData.mYSize.height);
+                       aData.mCbCrStride, ySize.width, ySize.height);
   } else {  // NV21
     libyuv::NV21ToI420(mData.mYChannel, mData.mYStride, mData.mCrChannel,
                        mData.mCbCrStride, aData.mYChannel, aData.mYStride,
                        aData.mCbChannel, aData.mCbCrStride, aData.mCrChannel,
-                       aData.mCbCrStride, aData.mYSize.width,
-                       aData.mYSize.height);
+                       aData.mCbCrStride, ySize.width, ySize.height);
   }
 
   // The logics in PlanarYCbCrImage::GetAsSourceSurface().
@@ -690,9 +893,6 @@ already_AddRefed<SourceSurface> NVImage::GetAsSourceSurface() {
 
   mSourceSurface = surface;
 
-  // Release the temporary buffer.
-  delete[] buffer;
-
   return surface.forget();
 }
 
@@ -709,8 +909,8 @@ bool NVImage::SetData(const Data& aData) {
   // Calculate buffer size
   // Use uint32_t throughout to match AllocateBuffer's param and mBufferSize
   const auto checkedSize =
-      CheckedInt<uint32_t>(aData.mYSize.height) * aData.mYStride +
-      CheckedInt<uint32_t>(aData.mCbCrSize.height) * aData.mCbCrStride;
+      CheckedInt<uint32_t>(aData.YDataSize().height) * aData.mYStride +
+      CheckedInt<uint32_t>(aData.CbCrDataSize().height) * aData.mCbCrStride;
 
   if (!checkedSize.isValid()) return false;
 
@@ -732,7 +932,7 @@ bool NVImage::SetData(const Data& aData) {
   mData.mCrChannel = mData.mYChannel + (aData.mCrChannel - aData.mYChannel);
 
   // Update mSize.
-  mSize = aData.mPicSize;
+  mSize = aData.mPictureRect.Size();
 
   // Copy the input data into mBuffer.
   // This copies the y-channel and the interleaving CbCr-channel.
@@ -743,55 +943,57 @@ bool NVImage::SetData(const Data& aData) {
 
 const NVImage::Data* NVImage::GetData() const { return &mData; }
 
-UniquePtr<uint8_t> NVImage::AllocateBuffer(uint32_t aSize) {
-  UniquePtr<uint8_t> buffer(new uint8_t[aSize]);
+UniquePtr<uint8_t[]> NVImage::AllocateBuffer(uint32_t aSize) {
+  UniquePtr<uint8_t[]> buffer(new uint8_t[aSize]);
   return buffer;
 }
 
 SourceSurfaceImage::SourceSurfaceImage(const gfx::IntSize& aSize,
                                        gfx::SourceSurface* aSourceSurface)
-    : Image(nullptr, ImageFormat::CAIRO_SURFACE),
+    : Image(nullptr, ImageFormat::MOZ2D_SURFACE),
       mSize(aSize),
       mSourceSurface(aSourceSurface),
       mTextureFlags(TextureFlags::DEFAULT) {}
 
 SourceSurfaceImage::SourceSurfaceImage(gfx::SourceSurface* aSourceSurface)
-    : Image(nullptr, ImageFormat::CAIRO_SURFACE),
+    : Image(nullptr, ImageFormat::MOZ2D_SURFACE),
       mSize(aSourceSurface->GetSize()),
       mSourceSurface(aSourceSurface),
       mTextureFlags(TextureFlags::DEFAULT) {}
 
-SourceSurfaceImage::~SourceSurfaceImage() = default;
+SourceSurfaceImage::~SourceSurfaceImage() {
+  NS_ReleaseOnMainThread("SourceSurfaceImage::mSourceSurface",
+                         mSourceSurface.forget());
+}
 
 TextureClient* SourceSurfaceImage::GetTextureClient(
-    KnowsCompositor* aForwarder) {
-  if (!aForwarder) {
+    KnowsCompositor* aKnowsCompositor) {
+  if (!aKnowsCompositor) {
     return nullptr;
   }
 
-  auto entry = mTextureClients.LookupForAdd(aForwarder->GetSerial());
-  if (entry) {
-    return entry.Data();
-  }
+  return mTextureClients.WithEntryHandle(
+      aKnowsCompositor->GetSerial(), [&](auto&& entry) -> TextureClient* {
+        if (entry) {
+          return entry->get();
+        }
 
-  RefPtr<TextureClient> textureClient;
-  RefPtr<SourceSurface> surface = GetAsSourceSurface();
-  MOZ_ASSERT(surface);
-  if (surface) {
-    // gfx::BackendType::NONE means default to content backend
-    textureClient = TextureClient::CreateFromSurface(
-        aForwarder, surface, BackendSelector::Content, mTextureFlags,
-        ALLOC_DEFAULT);
-  }
-  if (textureClient) {
-    textureClient->SyncWithObject(aForwarder->GetSyncObject());
-    entry.OrInsert([&textureClient]() { return textureClient; });
-    return textureClient;
-  }
+        RefPtr<TextureClient> textureClient;
+        RefPtr<SourceSurface> surface = GetAsSourceSurface();
+        MOZ_ASSERT(surface);
+        if (surface) {
+          // gfx::BackendType::NONE means default to content backend
+          textureClient = TextureClient::CreateFromSurface(
+              aKnowsCompositor, surface, BackendSelector::Content,
+              mTextureFlags, ALLOC_DEFAULT);
+        }
+        if (textureClient) {
+          textureClient->SyncWithObject(aKnowsCompositor->GetSyncObject());
+          return entry.Insert(std::move(textureClient)).get();
+        }
 
-  // Remove the speculatively added entry.
-  entry.OrRemove();
-  return nullptr;
+        return nullptr;
+      });
 }
 
 ImageContainer::ProducerID ImageContainer::AllocateProducerID() {
@@ -800,5 +1002,4 @@ ImageContainer::ProducerID ImageContainer::AllocateProducerID() {
   return ++sProducerID;
 }
 
-}  // namespace layers
-}  // namespace mozilla
+}  // namespace mozilla::layers
